
/*
 * sound/vivo.c
 *
 * Support routines for Ensoniq Soundscape VIVO. The actual driver is
 * in the vivo subdirectory.
 */
/*
 * Copyright (C) by Hannu Savolainen 1993-1996
 *
 * OSS/Free for Linux is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
 * Version 2 (June 1991). See the "COPYING" file distributed with this software
 * for more info.
 */

#include "sound_config.h"

extern void     otto_init (int options);
extern void     synth_init (void);
